package POWER::Event::Timer;

use version; $VERSION = qv('0.1.2'); # update POD and Changes!

use warnings;
use strict;
use Carp;

use List::Util      qw( min max );
use Scalar::Util    qw( weaken  );

use Perl6::Export::Attrs;
use List::MoreUtils qw( first_index );

use POWER::Utils;

my @Timers;             #   

sub _DUMP { # DEBUG
    warn "<<POWER::Event::Timer>>\n";
    for (@Timers) {
        warn sprintf "\tcb=%s\tmethod=%s\t%s\n",
            (defined $_->{cb} ? $_->{cb} : '(destroyed object)'),
            $_->{method},
            (exists $_->{for} ? "for=$_->{for}" : "every=$_->{every}"),
            ;
    }
}

sub set_timer :Export(:DEFAULT) {
    my ($opt_ref) = @_;
    croak 'usage: set_timer({ options })'   if ref $opt_ref ne 'HASH';
    croak 'set_timer: {cb} required'        if !exists $opt_ref->{cb};
    croak 'set_timer: one from {for} or {every} required'
        if !(exists $opt_ref->{for} xor exists $opt_ref->{every});
    croak "set_timer: {$_} undefined" for 
        grep {exists $opt_ref->{$_} && !defined $opt_ref->{$_}} qw( for every );
    croak "set_timer: wrong {for}=>$opt_ref->{for} (must be positive)"
        if exists $opt_ref->{for}   && $opt_ref->{for}   <= 0;
    croak "set_timer: wrong {every}=>$opt_ref->{every} (must be positive)"
        if exists $opt_ref->{every} && $opt_ref->{every} <= 0;
    croak "set_timer: unknown param {$_}"
        for grep { !m{\A(?:for|every|cb|method)\z}xms } keys %$opt_ref;
    my $timer = { method => 'ALARM', %$opt_ref };
    weaken $timer->{cb} if ref($timer->{cb}) && ref($timer->{cb}) ne 'CODE';
    reset_timer($timer);
    push @Timers, $timer;
    return $timer;
}

sub reset_timer :Export(:DEFAULT) {
    my ($timer) = @_;
    croak   'usage: reset_timer(TIMER)'    if ref $timer ne 'HASH';
    # BUG    reset_timer()   
    #      ţ   ̣ .
    $timer->{alarm_at} = clock_monotonic() 
        + (exists $timer->{for} ? $timer->{for} : $timer->{every});
    return;
}

sub stop_timer :Export(:DEFAULT) {
    my ($timer) = @_;
    croak q{usage: stop_timer(TIMER)} if ref $timer ne 'HASH';
    my $idx = first_index { $_ == $timer } @Timers; # compare refaddr's
    croak q{stop_timer: no such timer} if $idx == -1;
    splice @Timers, $idx, 1;
    return;
}

sub get_timers_idle_time :Export(:event) {
    return undef unless @Timers;
    my $nearest_alarm = min(map { $_->{alarm_at} } @Timers);
    return max(0, $nearest_alarm - clock_monotonic());
}

sub get_expired_timers :Export(:event) {
    my $clock = clock_monotonic();
    my @expired_timers;
    TIMER:
    for my $timer (splice @Timers) {
        next TIMER if !defined $timer->{cb};    # drop destroyed objects
        if ($timer->{alarm_at} > $clock) {
            push @Timers, $timer;
        }
        else {
            push @expired_timers, $timer;
            if (exists $timer->{every}) {
                reset_timer($timer);
                push @Timers, $timer;
            }
        }
    }
    return 
        map { ref($_->[0]) && ref($_->[0]) ne 'CODE' && weaken $_->[0]; $_ } 
        map { [$_->{cb}, $_->{method}, $_] }
        @expired_timers;
}

1; # Magic true value required at end of module
__END__

=head1 NAME

POWER::Event::Timer - Schedule callback execution after given timeout.


=head1 VERSION

This document describes POWER::Event::Timer version 0.1.2


=head1 SYNOPSIS

 # in user code:
 use POWER::Event::Timer;
 set_timer({ for => 2.5, cb => \&my_alarm });
 my $timer = set_timer({ every => 1, cb => 'MyCLASS' });
 reset_timer($timer);
 stop_timer($timer);

 # somethere in event loop:
 use POWER::Utils;                     # for run_callback()
 use POWER::Event::Timer qw( :event );
 my $idle_time = get_timers_idle_time();
 sleep $idle_time;
 for my $event (get_expired_timers()) {
     run_callback(@$event);
 }


=head1 DESCRIPTION

Allow you to:

 - set a lot of one-shot or repeated timers for callbacks which
   should be executed after some timeout
 - query how many time left before nearest timer will expire
 - fetch list of all callbacks with expired timers in format suitable
   for executing these callbacks with POWER::Utils/run_callback()
 - manage your timers (stop, reset)

All time should be given in floating seconds.

Interface of this module it developed for using in user's event loops,
for example see L<POWER::Event>.


=head1 EXPORTS

=over

=item :DEFAULT

set_timer() reset_timer() stop_timer()

=item :event

get_timers_idle_time() get_expired_timers()

=back

=head1 INTERFACE

=over

=item set_timer( OPTIONS )

Set up single-shot or repeated timer.

OPTIONS is a hash ref with possible keys:

 {for}      REQUIRED. Set up one-shot timer for {for} seconds.
 {every}    REQUIRED. Set up repeated timer every {every} seconds.
 {cb}       REQUIRED. Callback to execute when timer expires.
 {method}   OPTIONAL (default is 'ALARM').
 NOTE: {for} and {every} are mutually exclusive, so only
       one of them are really REQUIRED.
 NOTE: See POWER::Utils/run_callback() for details about
       {cb} and {method} format.

If {cb} is OBJECT then B<weaken> reference to OBJECT will be stored in
internal structures. If that OBJECT will be destroyed without calling
stop_timer() first, then get_expired_timers() I<will not return OBJECT>
in case it's timer(s) will expire. Also get_expired_timers() will automatically
stop_timer() all timers for destroyed OBJECTs, so you shouldn't manually
call stop_timer() I<after> OBJECT will be destroyed.

Return: timer identifier which can be used to manage this timer by
reset_timer() and stop_timer() .


=item reset_timer( TIMER )

Will reset time left before TIMER expiration to it {for} or {every} property.

TIMER is a value returned by set_timer().

Return: nothing.


=item stop_timer( TIMER )

Stop TIMER. After stop_timer() this TIMER can't be used any more.
You must not stop_timer() after single-shot timer expires - it will
be stopped automatically.

TIMER is a value returned by set_timer().

Return: nothing.


=item get_timers_idle_time( )

Return: time left before nearest timer will expire.

Useful to sleep() or do some operation with that timeout before execute
get_expired_timers().


=item get_expired_timers( )

Return: LIST with all expired timers in format suitable for
POWER::Utils/run_callback() (see L</SYNOPSIS> for usage example).

If timer doesn't have {every}, then stop this timer;
else reschedule this timer to expire after {every} seconds.


=back

=head1 DIAGNOSTICS

C<< ... >> in messages text below mean placeholder for some value.

=head2 EXCEPTIONS

=over

=item C<< usage: set_timer({ options }) >>

set_timer() was not called with hash reference in first param.

=item C<< set_timer: {cb} required >>

You don't give {cb} option to set_timer().

=item C<< set_timer: one from {for} or {every} required >>

You don't give {for} or {every} option to set_timer() OR you gived both
{for} and {every} at same time.

=item C<< set_timer: wrong {for}=>... (must be positive) >>

{for} must be > 0.

=item C<< set_timer: wrong {every}=>... (must be positive) >>

{every} must be > 0.

=item C<< set_timer: unknown param {...} >>

See list of supported params in set_timer() above .

=item C<< usage: reset_timer(TIMER) >>

reset_timer() was not called with value returned by set_timer() in first param.

=item C<< usage: stop_timer(TIMER) >>

stop_timer() was not called with value returned by set_timer() in first param.

=item C<< stop_timer: no such timer >>

Possible reasons:

 - this timer was already deleted by stop_timer()
 - this TIMER is single-shot and was already returned
   by get_expired_timers()

=back

=head2 WARNINGS

None.


=head1 CONFIGURATION AND ENVIRONMENT

POWER::Event::Timer requires no configuration files or environment variables.


=head1 DEPENDENCIES

 Perl6::Export::Attrs
 List::MoreUtils
 POWER::Utils


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

This version has no speed optimizations which probably will be required
for production use.

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-sub-scheduler@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.

=head1 AUTHOR

Alex Efros <powerman@asdfgroup.com>

Ilya Persky <ilya_persky@yahoo.com>

=head1 LICENCE AND COPYRIGHT

Copyright (c) 2005, Alex Efros <powerman@asdfgroup.com>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
